////////////////////////////////////////////////////////////////////////
// Class:       OnlineCluster
// Module Type: producer
// File:        OnlineCluster_module.cc
//
// Generated at Thu May  9 15:59:44 2013 by Nicola Rossi using artmod
// from cetpkgsupport v1_02_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "artdaq-core/Data/Fragments.hh"
#include "darksidecore/Data/V1495Fragment.hh"
#include "darksidecore/Data/V172xFragment.hh"
#include "darkmon/ArtModules/OnlineEvent.hh"

#include <numeric>
#include <vector>
#include <functional>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <math.h>
#include <fstream>
#include <string>

namespace {
  // moving average
  std::vector<float> moving_avg(std::vector<float>::const_iterator v_start, std::vector<float>::const_iterator v_stop, size_t m_size) {
    size_t v_size = v_stop - v_start;
    if(v_size < m_size) throw cet::exception("OnlineCluster") << "mowing avg larger than vector size";
    std::vector<float> v_out(v_size - m_size + 1, 0);
    v_out[0] = std::accumulate(v_start, v_start + m_size, 0.) / m_size;
    for (size_t i = 1; i < v_out.size(); i++) v_out[i] = (m_size * v_out[i - 1] + *(v_start + m_size - 1 + i) - *(v_start + i - 1)) / m_size;
    return v_out;
  }

  // stairize
  std::vector<float> stairize(std::vector<float>::const_iterator v_start, std::vector<float>::const_iterator v_stop, size_t st_size) {
    size_t v_size = v_stop - v_start;
    if(v_size < st_size) throw cet::exception("OnlineCluster") << "stairs larger than vector size";
    std::vector<float> v_out(v_size / st_size , 0);
    for(size_t i = 0; i < v_out.size(); i++) v_out[i] = std::accumulate(v_start + i * st_size, v_start + (i + 1) *  st_size, 0.) / st_size;
    return v_out;
  }

  // peak finder
  std::vector<ds50::Online::Peak> peak_finder(std::vector<float>::const_iterator v_start, std::vector<float>::const_iterator v_stop, std::vector<float>::const_iterator vr_start, std::vector<float>::const_iterator vr_stop, int c_time, size_t st_size, float gap, int t_sample, int step){
    std::vector<ds50::Online::Peak> peaks;
    ds50::Online::Peak c_peak;
    std::vector<float> v0 = moving_avg(v_start, v_stop, st_size);
    std::vector<float> v_tmp = stairize(v0.begin(), v0.end(), st_size);
    std::vector<float> dv(v_tmp.size());
    std::adjacent_difference(v_tmp.begin(), v_tmp.end(), dv.begin());
    for (size_t j = 0; j < dv.size() - 1; j++ ){
      if(dv[j] == dv[j + 1]) continue;
      if(dv[j] > 0 && dv[j + 1] < 0 && dv[j] - dv[j + 1] > gap) {
        int t_max = std::distance(vr_start, std::max_element(vr_start + (j - 1) * st_size, vr_start  + (j + 1) * st_size));
        c_peak.time_ns = t_sample * (c_time + step - int(vr_stop - vr_start) + t_max);
        c_peak.height = *(vr_start + t_max);
        peaks.push_back(c_peak);
      }
    }
    if (!peaks.size()) {
      int t_max = std::distance(vr_start, std::max_element(vr_start , vr_stop));
      c_peak.time_ns = t_sample * (t_max + step + c_time - int(vr_stop - vr_start));
      c_peak.height = *(vr_start + t_max);
      peaks.push_back(c_peak);
    }
    return peaks;
  }

  // charge
  float fcharge(std::vector<float>::const_iterator &v, size_t cluster_size) {
    return std::accumulate(v, v + cluster_size, 0.); 
  }

  // fabs_charge
  float fabs_charge(std::vector<float>::const_iterator &v, size_t cluster_size) {
    float fen = 0.;
    for (size_t i = 0; i < cluster_size; i++) fen += std::fabs(*(v + i));
    return fen;
  }

  // psd
  float psd_f(std::vector<float>::const_iterator &v, size_t cluster_size, int f_stop) {
    return std::accumulate (v, v + f_stop, 0.) / std::accumulate(v, v + cluster_size, 0.); 
  }

  // mean time
  float meantime(std::vector<float>::const_iterator &v, size_t cluster_size, int t_sample) {
    float fcharge = fabs_charge (v, cluster_size);
    if (!fcharge) return 0;

    float mt = 0;
    for (size_t i = 0; i < cluster_size; i++) mt += std::fabs (*(v + i)) * (i * t_sample);

    return mt / fcharge;
  }

  // rms
  float rms(std::vector<float>::const_iterator &v, size_t cluster_size, int t_sample) {
    float fcharge = fabs_charge (v, cluster_size);
    if (!fcharge) return 0;

    float rms_t = 0, mt = meantime (v, cluster_size, t_sample);
    for (size_t i = 0; i < cluster_size; i++) rms_t += std::fabs (*(v + i)) * (i * t_sample - mt) * (i * t_sample - mt);

    return std::sqrt(rms_t / std::fabs(fcharge));
  }
}

namespace ds50 {
  class OnlineCluster : public art::EDProducer {
    public:
      explicit OnlineCluster(fhicl::ParameterSet const & p);
      virtual ~OnlineCluster();

      void produce(art::Event & e) override;
      void beginRun(art::Run &) override;

    private:
      std::string board_type_;
      size_t mavg_;
      int stairs_;
      int gwidth_us_;
      float delta_;
      float threshold_;
      std::vector<float> x_;
  };
}

ds50::OnlineCluster::OnlineCluster(fhicl::ParameterSet const & ps): board_type_(ps.get<std::string> ("board_type", "V1720")), mavg_(ps.get<size_t>("mov_avg")), stairs_(ps.get<int>("stairs")), gwidth_us_(ps.get<int>("gwidth_us")), delta_(ps.get<float>("delta")), threshold_(ps.get<float>("threshold")) {
   produces<ds50::Online::Clusters> ();
 }

ds50::OnlineCluster::~OnlineCluster() {}

void ds50::OnlineCluster::produce(art::Event & e) {
  art::Handle<artdaq::Fragments> v1495;
  e.getByLabel ("daq", "V1495", v1495);
  if (!v1495.isValid ()) return;

  if(v1495->size() != 1) {
    mf::LogError ("OnlineCluster") << v1495->size() << " fragment for V1495";
    return;
  }

  if (V1495Fragment((*v1495)[0]).trigger_type () != V1495Fragment::tpc) return;

  art::Handle<ds50::Online::Waveforms> wf;
  e.getByLabel ("waveform", wf);
  if (!wf.isValid ()) return;

  const std::vector<float>& y_sum = wf->sum;

  std::vector<float> y_m = moving_avg(y_sum.begin(), y_sum.end(), mavg_);

  std::unique_ptr<ds50::Online::Clusters> clusters(new ds50::Online::Clusters);
  clusters->gps_time = V1495Fragment((*v1495)[0]).gps_coarse_s () + V1495Fragment((*v1495)[0]).gps_fine_20ns () * 20e-9;

  int gwidth = gwidth_us_ * 1e3 / wf->sample_ns;
  for (size_t i = stairs_ ; i < y_m.size() - gwidth - stairs_; i++){
    if (y_m[i] < threshold_) continue;
    std::vector<float> y_temp_c(0);
    int c_rep = 0;
    while (i < y_m.size() - gwidth - stairs_) {
      i += gwidth;
      c_rep ++;
      if (y_m[i] < threshold_ && std::accumulate(y_m.begin() + i - stairs_, y_m.begin() + i + stairs_, 0.) / 2 / stairs_ < threshold_ * mavg_ / stairs_ ) break;
    }

    ds50::Online::Cluster current_cluster;

    std::vector<float>::const_iterator c_start = y_sum.begin() + i - c_rep * gwidth + int(mavg_ / 2);
    size_t c_size = c_rep * gwidth;

    current_cluster.start_time_ns = (std::distance (std::vector<float>::const_iterator(y_sum.begin()), c_start) - wf->pre_samples)  * wf->sample_ns;
    current_cluster.stop_time_ns = current_cluster.start_time_ns + c_size * wf->sample_ns;
    current_cluster.mean_time_ns = meantime(c_start, c_size, wf->sample_ns);
    current_cluster.rms_time_ns = rms(c_start, c_size, wf->sample_ns);
    current_cluster.cluster_rep = c_rep;
    current_cluster.charge = fcharge(c_start, c_size);
    current_cluster.abs_charge = fabs_charge(c_start, c_size);
    current_cluster.psd[ds50::Online::Cluster::f80] = psd_f(c_start, c_size, 80 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f85] = psd_f(c_start, c_size, 85 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f90] = psd_f(c_start, c_size, 90 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f95] = psd_f(c_start, c_size, 95 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f100] = psd_f(c_start, c_size, 100 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f105] = psd_f(c_start, c_size, 105 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f110] = psd_f(c_start, c_size, 110 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f115] = psd_f(c_start, c_size, 115 / wf->sample_ns);
    current_cluster.psd[ds50::Online::Cluster::f120] = psd_f(c_start, c_size, 120 / wf->sample_ns);
    for (int k = 0; k < 40; k++) 
      if (wf->wf[k].size () > 0) {
	c_start = wf->wf[k].begin () + i - c_rep * gwidth + int(mavg_ / 2);
	current_cluster.pmts[k].charge = fcharge (c_start, c_size);
	current_cluster.pmts[k].abs_charge = fabs_charge (c_start, c_size);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f80] = psd_f(c_start, c_size, 80 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f85] = psd_f(c_start, c_size, 85 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f90] = psd_f(c_start, c_size, 90 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f95] = psd_f(c_start, c_size, 95 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f100] = psd_f(c_start, c_size, 100 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f105] = psd_f(c_start, c_size, 105 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f110] = psd_f(c_start, c_size, 110 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f115] = psd_f(c_start, c_size, 115 / wf->sample_ns);
	current_cluster.pmts[k].psd[ds50::Online::Cluster::f120] = psd_f(c_start, c_size, 120 / wf->sample_ns);
      } else {
	current_cluster.pmts[k].charge = current_cluster.pmts[k].abs_charge = -100;
	std::fill (current_cluster.pmts[k].psd, current_cluster.pmts[k].psd + sizeof (current_cluster.pmts[k].psd) / sizeof (*current_cluster.pmts[k].psd), -100);
      }

    //peak finder
    current_cluster.peaks = peak_finder(y_m.begin() + i  - c_rep * gwidth, y_m.begin() + i , y_sum.begin() + i - c_rep * gwidth + int(mavg_ / 2) , y_sum.begin() + i + int(mavg_ / 2) , i, stairs_, delta_, wf->sample_ns, int(mavg_ / 2));
    for (size_t p = 0; p < current_cluster.peaks.size(); p++) current_cluster.peaks[p].time_ns -= current_cluster.start_time_ns + wf->pre_samples * wf->sample_ns;
    clusters->clusters.push_back (current_cluster);
  }

  e.put(std::move(clusters));
}

void ds50::OnlineCluster::beginRun(art::Run &) {}

DEFINE_ART_MODULE(ds50::OnlineCluster)
